from .dataset import IdxDataset, NpzDataset


class MNIST(IdxDataset):
    """
    MNIST Dataset
    http://yann.lecun.com/exdb/mnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    classes : list[str]
        Class names.
    mirrors : list[str]
        List of urls where dataset is hosted.
    resources : dict[str, tuple[str, str]]
       Dictionary of data files with filename and md5 hash.

    Usage
    -----
    >>> from mnists import MNIST
    >>> mnist = MNIST()
    >>> type(mnist.train_images())
    <class 'numpy.ndarray'>
    >>> mnist.train_images().dtype
    dtype('uint8')
    >>> mnist.train_images().min()
    0
    >>> mnist.train_images().max()
    255
    >>> mnist.train_images().shape
    (60000, 28, 28)
    >>> mnist.train_labels().shape
    (60000,)
    >>> mnist.test_images().shape
    (10000, 28, 28)
    >>> mnist.test_labels().shape
    (10000,)

    Citation
    --------
    @article{lecun-98,
      author={Lecun, Y. and Bottou, L. and Bengio, Y. and Haffner, P.},
      journal={Proceedings of the IEEE},
      title={Gradient-based learning applied to document recognition},
      year={1998},
      volume={86},
      number={11},
      pages={2278-2324},
      doi={10.1109/5.726791}
    }
    """

    classes = [
        "0 - zero",
        "1 - one",
        "2 - two",
        "3 - three",
        "4 - four",
        "5 - five",
        "6 - six",
        "7 - seven",
        "8 - eight",
        "9 - nine",
    ]

    mirrors = [
        "https://storage.googleapis.com/cvdf-datasets/mnist/",
        "https://ossci-datasets.s3.amazonaws.com/mnist/",
        "http://yann.lecun.com/exdb/mnist/",
    ]

    resources = {
        "train_images": (
            "train-images-idx3-ubyte.gz",
            "f68b3c2dcbeaaa9fbdd348bbdeb94873",
        ),
        "train_labels": (
            "train-labels-idx1-ubyte.gz",
            "d53e105ee54ea40749a09fcbcd1e9432",
        ),
        "test_images": (
            "t10k-images-idx3-ubyte.gz",
            "9fb629c4189551a2d022fa330f9573f3",
        ),
        "test_labels": (
            "t10k-labels-idx1-ubyte.gz",
            "ec29112dd5afa0611ce80d1b7f02629c",
        ),
    }


class FashionMNIST(IdxDataset):
    """
    Fashion-MNIST Dataset
    https://github.com/zalandoresearch/fashion-mnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    classes : list[str]
        Class names.
    mirrors : list[str]
        List of urls where dataset is hosted.
    resources : dict[str, tuple[str, str]]
       Dictionary of data files with filename and md5 hash.

    Usage
    -----
    >>> from mnists import FashionMNIST
    >>> fmnist = FashionMNIST()
    >>> fmnist.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @online{xiao2017/online,
      author       = {Han Xiao and Kashif Rasul and Roland Vollgraf},
      title        = {Fashion-MNIST: a Novel Image Dataset for Benchmarking
                      Machine Learning Algorithms},
      date         = {2017-08-28},
      year         = {2017},
      eprintclass  = {cs.LG},
      eprinttype   = {arXiv},
      eprint       = {cs.LG/1708.07747},
    }
    """

    classes = [
        "T-shirt/top",
        "Trouser",
        "Pullover",
        "Dress",
        "Coat",
        "Sandal",
        "Shirt",
        "Sneaker",
        "Bag",
        "Ankle boot",
    ]

    mirrors = [
        "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/",
    ]

    resources = {
        "train_images": (
            "train-images-idx3-ubyte.gz",
            "8d4fb7e6c68d591d4c3dfef9ec88bf0d",
        ),
        "train_labels": (
            "train-labels-idx1-ubyte.gz",
            "25c81989df183df01b3e8a0aad5dffbe",
        ),
        "test_images": (
            "t10k-images-idx3-ubyte.gz",
            "bef4ecab320f06d8554ea6380940ec79",
        ),
        "test_labels": (
            "t10k-labels-idx1-ubyte.gz",
            "bb300cfdad3c16e7a12a480ee83cd310",
        ),
    }


class KMNIST(IdxDataset):
    """
    Kuzushiji-MNIST Dataset
    https://github.com/rois-codh/kmnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    classes : list[str]
        Class names.
    mirrors : list[str]
        List of urls where dataset is hosted.
    resources : dict[str, tuple[str, str]]
       Dictionary of data files with filename and md5 hash.

    Usage
    -----
    >>> from mnists import KMNIST
    >>> kmnist = KMNIST()
    >>> kmnist.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @online{clanuwat2018deep,
      author       = {Tarin Clanuwat and Mikel Bober-Irizar and Asanobu Kitamoto
                      and Alex Lamb and Kazuaki Yamamoto and David Ha},
      title        = {Deep Learning for Classical Japanese Literature},
      date         = {2018-12-03},
      year         = {2018},
      eprintclass  = {cs.CV},
      eprinttype   = {arXiv},
      eprint       = {cs.CV/1812.01718},
    }
    """

    classes = [
        "お - o",
        "き - ki",
        "す - su",
        "つ - tsu",
        "な - na",
        "は - ha",
        "ま - ma",
        "や - ya",
        "れ - re",
        "を - wo",
    ]

    mirrors = [
        "http://codh.rois.ac.jp/kmnist/dataset/kmnist/",
    ]

    resources = {
        "train_images": (
            "train-images-idx3-ubyte.gz",
            "bdb82020997e1d708af4cf47b453dcf7",
        ),
        "train_labels": (
            "train-labels-idx1-ubyte.gz",
            "e144d726b3acfaa3e44228e80efcd344",
        ),
        "test_images": (
            "t10k-images-idx3-ubyte.gz",
            "5c965bf0a639b31b8f53240b1b52f4d7",
        ),
        "test_labels": (
            "t10k-labels-idx1-ubyte.gz",
            "7320c461ea6c1c855c0b718fb2a4b134",
        ),
    }


class K49(NpzDataset):
    """
    Kuzushiji-49 Dataset
    https://github.com/rois-codh/kmnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    classes : list[str]
        Class names.
    mirrors : list[str]
        List of urls where dataset is hosted.
    resources : dict[str, tuple[str, str]]
       Dictionary of data files with filename and md5 hash.

    Usage
    -----
    >>> from mnists import K49
    >>> k49 = K49()
    >>> k49.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @online{clanuwat2018deep,
      author       = {Tarin Clanuwat and Mikel Bober-Irizar and Asanobu Kitamoto
                      and Alex Lamb and Kazuaki Yamamoto and David Ha},
      title        = {Deep Learning for Classical Japanese Literature},
      date         = {2018-12-03},
      year         = {2018},
      eprintclass  = {cs.CV},
      eprinttype   = {arXiv},
      eprint       = {cs.CV/1812.01718},
    }
    """

    classes = [
        "あ - a",
        "い - i",
        "う - u",
        "え - e",
        "お - o",
        "か - ka",
        "き - ki",
        "く - ku",
        "け - ke",
        "こ - ko",
        "さ - sa",
        "し - shi",
        "す - su",
        "せ - se",
        "そ - so",
        "た - ta",
        "ち - chi",
        "つ - tsu",
        "て - te",
        "と - to",
        "な - na",
        "に - ni",
        "ぬ - nu",
        "ね - ne",
        "の - no",
        "は - ha",
        "ひ - hi",
        "ふ - fu",
        "へ - he",
        "ほ - ho",
        "ま - ma",
        "み - mi",
        "む - mu",
        "め - me",
        "も - mo",
        "や - ya",
        "ゆ - yu",
        "よ - yo",
        "ら - ra",
        "り - ri",
        "る - ru",
        "れ - re",
        "ろ - ro",
        "わ - wa",
        "ゐ - i",
        "ゑ - e",
        "を - wo",
        "ん - n",
        "ゝ - iteration mark",
    ]

    mirrors = [
        "http://codh.rois.ac.jp/kmnist/dataset/k49/",
    ]

    resources = {
        "train_images": (
            "k49-train-imgs.npz",
            "7ac088b20481cf51dcd01ceaab89d821",
        ),
        "train_labels": (
            "k49-train-labels.npz",
            "44a8e1b893f81e63ff38d73cad420f7a",
        ),
        "test_images": (
            "k49-test-imgs.npz",
            "d352e201d846ce6b94f42c990966f374",
        ),
        "test_labels": (
            "k49-test-labels.npz",
            "4da6f7a62e67a832d5eb1bd85c5ee448",
        ),
    }



from typing import Optional

from .dataset import SplitDataset, ZippedDataset


class EMNIST(SplitDataset):
    """
    EMNIST Dataset
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    Balanced, ByClass, ByMerge, Digits, Letters : class
        Child classes containing splits of EMNIST dataset.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> letters = emnist.Letters()
    >>> letters.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    mirrors = [
        "https://biometrics.nist.gov/cs_links/EMNIST/",
    ]

    resources = {"gzip": ("gzip.zip", "58c8d27c78d21e728a6bc7b3cc06412e")}

    def __init__(
        self,
        target_dir: Optional[str] = None,
        download: bool = True,
        force_download: bool = False,
    ) -> None:
        super().__init__(target_dir, download, force_download)

        self.Balanced = self._create_split(Balanced)
        self.ByClass = self._create_split(ByClass)
        self.ByMerge = self._create_split(ByMerge)
        self.Digits = self._create_split(Digits)
        self.Letters = self._create_split(Letters)


class Balanced(ZippedDataset):
    """
    EMNIST Balanced
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    zip_filepath : str
        Zip file from which dataset will be extracted.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> balanced = emnist.Balanced()
    >>> balanced.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    classes = [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "A",
        "B",
        "C - c",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I - i",
        "J - j",
        "K - k",
        "L - l",
        "M - m",
        "N",
        "O - o",
        "P - p",
        "Q",
        "R",
        "S - s",
        "T",
        "U - u",
        "V - v",
        "W - w",
        "X - x",
        "Y - y",
        "Z - z",
        "a",
        "b",
        "d",
        "e",
        "f",
        "g",
        "h",
        "n",
        "q",
        "r",
        "t",
    ]

    resources = {
        "train_images": (
            "emnist-balanced-train-images-idx3-ubyte.gz",
            "4041b0d6f15785d3fa35263901b5496b",
        ),
        "train_labels": (
            "emnist-balanced-train-labels-idx1-ubyte.gz",
            "7a35cc7b2b7ee7671eddf028570fbd20",
        ),
        "test_images": (
            "emnist-balanced-test-images-idx3-ubyte.gz",
            "6818d20fe2ce1880476f747bbc80b22b",
        ),
        "test_labels": (
            "emnist-balanced-test-labels-idx1-ubyte.gz",
            "acd3694070dcbf620e36670519d4b32f",
        ),
    }


class ByClass(ZippedDataset):
    """
    EMNIST ByClass
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    zip_filepath : str
        Zip file from which dataset will be extracted.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> byclass = emnist.ByClass()
    >>> byclass.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    classes = [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
    ]

    resources = {
        "train_images": (
            "emnist-byclass-train-images-idx3-ubyte.gz",
            "712dda0bd6f00690f32236ae4325c377",
        ),
        "train_labels": (
            "emnist-byclass-train-labels-idx1-ubyte.gz",
            "ee299a3ee5faf5c31e9406763eae7e43",
        ),
        "test_images": (
            "emnist-byclass-test-images-idx3-ubyte.gz",
            "1435209e34070a9002867a9ab50160d7",
        ),
        "test_labels": (
            "emnist-byclass-test-labels-idx1-ubyte.gz",
            "7a0f934bd176c798ecba96b36fda6657",
        ),
    }


class ByMerge(ZippedDataset):
    """
    EMNIST ByMerge
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    zip_filepath : str
        Zip file from which dataset will be extracted.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> bymerge = emnist.ByMerge()
    >>> bymerge.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    resources = {
        "train_images": (
            "emnist-bymerge-train-images-idx3-ubyte.gz",
            "4a792d4df261d7e1ba27979573bf53f3",
        ),
        "train_labels": (
            "emnist-bymerge-train-labels-idx1-ubyte.gz",
            "491be69ef99e1ab1f5b7f9ccc908bb26",
        ),
        "test_images": (
            "emnist-bymerge-test-images-idx3-ubyte.gz",
            "8eb5d34c91f1759a55831c37ec2a283f",
        ),
        "test_labels": (
            "emnist-bymerge-test-labels-idx1-ubyte.gz",
            "c13f4cd5211cdba1b8fa992dae2be992",
        ),
    }


class Digits(ZippedDataset):
    """
    EMNIST Digits
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    zip_filepath : str
        Zip file from which dataset will be extracted.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> digits = emnist.Digits()
    >>> digits.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    classes = [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
    ]

    resources = {
        "train_images": (
            "emnist-digits-train-images-idx3-ubyte.gz",
            "d2662ecdc47895a6bbfce25de9e9a677",
        ),
        "train_labels": (
            "emnist-digits-train-labels-idx1-ubyte.gz",
            "2223fcfee618ac9c89ef20b6e48bcf9e",
        ),
        "test_images": (
            "emnist-digits-test-images-idx3-ubyte.gz",
            "a159b8b3bd6ab4ed4793c1cb71a2f5cc",
        ),
        "test_labels": (
            "emnist-digits-test-labels-idx1-ubyte.gz",
            "8afde66ea51d865689083ba6bb779fac",
        ),
    }


class Letters(ZippedDataset):
    """
    EMNIST Letters
    https://www.westernsydney.edu.au/bens/home/reproducible_research/emnist

    Attributes
    ----------
    target_dir : str
        Directory where all files exist or will be downloaded.
    zip_filepath : str
        Zip file from which dataset will be extracted.

    Usage
    -----
    >>> from mnists import EMNIST
    >>> emnist = EMNIST()
    >>> letters = emnist.Letters()
    >>> letters.train_images().dtype
    dtype('uint8')

    Citation
    --------
    @article{cohen2017emnist,
      title={EMNIST: an extension of MNIST to handwritten letters},
      author={Gregory Cohen and Saeed Afshar and Jonathan Tapson and André van Schaik},
      year={2017},
      eprint={1702.05373},
      archivePrefix={arXiv},
      primaryClass={cs.CV}
    }
    """

    classes = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
    ]

    resources = {
        "train_images": (
            "emnist-letters-train-images-idx3-ubyte.gz",
            "8795078f199c478165fe18db82625747",
        ),
        "train_labels": (
            "emnist-letters-train-labels-idx1-ubyte.gz",
            "c16de4f1848ddcdddd39ab65d2a7be52",
        ),
        "test_images": (
            "emnist-letters-test-images-idx3-ubyte.gz",
            "382093a19703f68edac6d01b8dfdfcad",
        ),
        "test_labels": (
            "emnist-letters-test-labels-idx1-ubyte.gz",
            "d4108920cd86601ec7689a97f2de7f59",
        ),
    }



import gzip
import hashlib
import os
import struct
import time
import zipfile
from urllib.error import URLError
from urllib.parse import urljoin
from urllib.request import urlretrieve

import numpy as np

_TQDM_ACTIVE = True
try:
    from tqdm import tqdm
except ImportError:
    tqdm = object
    _TQDM_ACTIVE = False


IDX_TYPEMAP = {
    0x08: np.uint8,
    0x09: np.int8,
    0x0B: np.int16,
    0x0C: np.int32,
    0x0D: np.float32,
    0x0E: np.float64,
}


def read_idx_file(filepath: str) -> np.ndarray:
    """
    Read file in IDX format and return numpy array.

    Parameters
    ----------
    filepath : str
        Path to a IDX file. The file can be gzipped.

    Returns
    -------
    np.ndarray
        Data read from IDX file in numpy array.
    """

    fopen = gzip.open if os.path.splitext(filepath)[1] == ".gz" else open

    with fopen(filepath, "rb") as f:
        data = f.read()

    h_len = 4
    header = data[:h_len]
    zeros, dtype, ndims = struct.unpack(">HBB", header)

    if zeros != 0:
        raise RuntimeError(
            "Invalid IDX file, file must start with two zero bytes. "
            f"Found 0x{zeros:X}"
        )

    try:
        dtype = IDX_TYPEMAP[dtype]
    except KeyError as e:
        raise RuntimeError(f"Unknown data type 0x{dtype:02X} in IDX file") from e

    dim_offset = h_len
    dim_len = 4 * ndims
    dim_sizes = data[dim_offset : dim_offset + dim_len]
    dim_sizes = struct.unpack(">" + "I" * ndims, dim_sizes)

    data_offset = h_len + dim_len
    parsed = np.frombuffer(data, dtype=dtype, offset=data_offset)

    if parsed.shape[0] != np.prod(dim_sizes):
        raise RuntimeError(
            f"Declared size {dim_sizes}={np.prod(dim_sizes)} and "
            f"actual size {parsed.shape[0]} of data in IDX file don't match"
        )

    return parsed.reshape(dim_sizes)


def check_file_integrity(filepath: str, md5: str) -> bool:
    """
    Check if file exists and if exists if its MD5 checksum is correct.

    Parameters
    ----------
    filepath : str
        Path to a file.
    md5 : str
        Correct MD5 checksum of the file.

    Returns
    -------
    bool
        Returns True when file exists and its MD5 checksum is equal `md5`.
    """

    return os.path.isfile(filepath) and md5 == calculate_md5(filepath)


def calculate_md5(filepath: str, chunk_size: int = 1024 * 1024) -> str:
    """
    Calculate MD5 checksum of the file.

    Parameters
    ----------
    filepath : str
        Path to a file.
    chunk_size : int, default=1024 * 1024
        Size of chunks which will be read from the file.

    Returns
    -------
    str
        MD5 checksum of the file.
    """

    md5 = hashlib.md5()
    with open(filepath, "rb") as fd:
        while chunk := fd.read(chunk_size):
            md5.update(chunk)
    return md5.hexdigest()


class EmptyTqdm(object):
    # https://github.com/tensorflow/datasets/blob/master/tensorflow_datasets/core/utils/tqdm_utils.py#L56
    def __init__(self, *args, **kwargs):
        self._iterator = args[0] if args else None

    def __iter__(self):
        return iter(self._iterator)

    def __getattr__(self, _):
        def empty_fn(*args, **kwargs):
            return

        return empty_fn

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        return


class Tqdm(tqdm):
    # https://github.com/tqdm/tqdm/blob/master/examples/tqdm_wget.py
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)


def custom_tqdm(*args, **kwargs):
    if _TQDM_ACTIVE:
        return Tqdm(*args, **kwargs)
    else:
        return EmptyTqdm(*args, **kwargs)


def download_file(mirrors: list[str], filename: str, filepath: str) -> None:
    """
    Download file trying every mirror if the previous one fails.

    Parameters
    ----------
    mirrors : list[str]
        List of the URLs of the mirrors.
    filename: str
        Name of the file on the server.
    filepath : str
        Path to the output file.
    """

    for mirror in mirrors:
        url = urljoin(mirror, filename)
        try:
            print(f"Downloading {url} to {filepath}")
            with custom_tqdm(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                miniters=1,
                desc=filepath,
            ) as t:
                urlretrieve(url, filepath, reporthook=t.update_to)
                t.total = t.n
            return
        except URLError as error:
            print(f"Failed to download {url} (trying next mirror):\n{error}")
            continue

    raise RuntimeError(f"Error downloading {filename}")


def extract_from_zip(zip_path: str, filename: str, output_dir: str) -> None:
    """
    Extract file from zip and save it to given directory (with correct metadata).

    Parameters
    ----------
    zip_path : str
        Path to the zip archive.
    filename : str
        Name of the file to be extracted.
    output_dir : str
        Directory where the file will be saved.
    """

    with zipfile.ZipFile(zip_path, "r") as archive:
        file = list(
            filter(
                lambda s: os.path.basename(s.filename) == filename, archive.infolist()
            )
        )

        if len(file) != 1:
            raise RuntimeError(
                f"Error while extracting {filename}: "
                f"found {len(file)} corresponding files in {zip_path}"
            )

        file = file[0]

        file.filename = os.path.basename(file.filename)
        archive.extract(file, output_dir)

        # add correct datetime metadata
        date_time = time.mktime(file.date_time + (0, 0, -1))
        os.utime(os.path.join(output_dir, file.filename), (date_time, date_time))



"""
MNISTs
======

Provides an easy access to MNIST-like datasets in a numpy format:
  1. MNIST
  2. FashionMNIST, FMNIST
  3. KMNIST, KuzushijiMNIST
  4. EMNIST - divided into Balanced, ByClass, ByMerge, Digits and Letters
     subsets
  5. Kuzushiji49, K49

Every dataset contains four numpy arrays:
  1. ``train_images`` of size ``(n_train_samples, width, height)``
  1. ``train_labels`` of size ``(n_train_samples,)``
  1. ``test_images`` of size ``(n_test_samples, width, height)``
  1. ``test_labels`` of size ``(n_test_samples,)``

All arrays are of type ``uint8``.

Example usage
-------------

  >>> from mnists import MNIST
  >>> mnist = MNIST()
  >>> type(mnist.train_images())
  <class 'numpy.ndarray'>
  >>> mnist.train_images().dtype
  dtype('uint8')
  >>> mnist.train_images().min()
  0
  >>> mnist.train_images().max()
  255
  >>> mnist.train_images().shape
  (60000, 28, 28)
  >>> mnist.train_labels().shape
  (60000,)
  >>> mnist.test_images().shape
  (10000, 28, 28)
  >>> mnist.test_labels().shape
  (10000,)

"""

from ._emnist import EMNIST
from ._mnist import K49, KMNIST, MNIST, FashionMNIST

FMNIST = FashionMNIST
KuzushijiMNIST = KMNIST
Kuzushiji49 = K49

__version__ = "0.4.0"



import os
import tempfile
from typing import Optional

import numpy as np

from .utils import check_file_integrity, download_file, extract_from_zip, read_idx_file

TEMPORARY_DIR = os.path.join(tempfile.gettempdir(), "mnists")


class Dataset:
    mirrors = []
    resources = {}

    def __init__(
        self,
        target_dir: Optional[str] = None,
        download: bool = True,
        force_download: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        target_dir : str, default='/tmp/<dataset_name>/'
            Directory where all files exist or will be downloaded to (if `download` is True).
        download : bool, default=True
            If True and files don't exist in `target_dir`, downloads all files to `target_dir`.
        force_download : bool, default=False
            If True, downloads all files to `target_dir`, even if they exist there.
        """

        self.target_dir = (
            os.path.join(TEMPORARY_DIR, type(self).__name__)
            if target_dir is None
            else target_dir
        )

        if download or force_download:
            self.download(force_download)

    def download(self, force: bool = False) -> None:
        """
        Download files from mirrors and save to `target_dir`.

        Parameters
        ----------
        force : bool=False
            If True, downloads all files even if they exist.
        """

        os.makedirs(self.target_dir, exist_ok=True)

        for filename, md5 in self.resources.values():
            filepath = os.path.join(self.target_dir, filename)

            if not force and check_file_integrity(filepath, md5):
                continue

            download_file(self.mirrors, filename, filepath)


class IdxDataset(Dataset):
    def __init__(
        self,
        target_dir: Optional[str] = None,
        download: bool = True,
        force_download: bool = False,
        load: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        target_dir : str, default='/tmp/<dataset_name>/'
            Directory where all files exist or will be downloaded to (if `download` is True).
        download : bool, default=True
            If True and files don't exist in `target_dir`, downloads all files to `target_dir`.
        force_download : bool, default=False
            If True, downloads all files to `target_dir`, even if they exist there.
        load : bool, default=True
            If True, loads data from files in `target_dir`.
        """

        self.target_dir = (
            os.path.join(TEMPORARY_DIR, type(self).__name__)
            if target_dir is None
            else target_dir
        )

        self._train_images: Optional[np.ndarray] = None
        self._train_labels: Optional[np.ndarray] = None
        self._test_images: Optional[np.ndarray] = None
        self._test_labels: Optional[np.ndarray] = None

        if download or force_download:
            self.download(force_download)

        if load:
            self.load()

    def train_images(self) -> np.ndarray:
        """
        Return train_images numpy array.

        Returns
        -------
        np.ndarray
        """
        if self._train_images is None:
            self._raise_dataset_not_loaded()
        return self._train_images

    def train_labels(self) -> np.ndarray:
        """
        Return train_labels numpy array.

        Returns
        -------
        np.ndarray
        """
        if self._train_labels is None:
            self._raise_dataset_not_loaded()
        return self._train_labels

    def test_images(self) -> np.ndarray:
        """
        Return test_images numpy array.

        Returns
        -------
        np.ndarray
        """
        if self._test_images is None:
            self._raise_dataset_not_loaded()
        return self._test_images

    def test_labels(self) -> np.ndarray:
        """
        Return test_labels numpy array.

        Returns
        -------
        np.ndarray
        """
        if self._test_labels is None:
            self._raise_dataset_not_loaded()
        return self._test_labels

    def _raise_dataset_not_loaded(self):
        raise RuntimeError(
            "Dataset wasn't loaded. You need to run .load() or create new "
            "object with load=True"
        )

    def load(self, transpose=False) -> None:
        """
        Load data from files in `target_dir`.

        Parameters
        ----------
        transpose : bool=False
            If True, transposes train and test images.
        """

        for key, (filename, md5) in self.resources.items():
            filepath = os.path.join(self.target_dir, filename)

            if not check_file_integrity(filepath, md5):
                raise RuntimeError(
                    f"Dataset '{key}' not found in '{filepath}' or MD5 "
                    "checksum is not valid. "
                    "Use download=True or .download() to download it"
                )

            data = read_idx_file(filepath)
            setattr(self, f"_{key}", data)

        if transpose:
            self._transpose_images()

    def _transpose_images(self) -> None:
        self._train_images = np.moveaxis(self._train_images, -2, -1)
        self._test_images = np.moveaxis(self._test_images, -2, -1)


class SplitDataset(Dataset):
    resources = {"gzip": (None, None)}

    def _create_split(self, split_cls: type["ZippedDataset"]) -> type["ZippedDataset"]:
        split_cls.default_base_dir = self.target_dir
        file, md5 = self.resources["gzip"]
        split_cls.default_zip_filepath = os.path.join(self.target_dir, file)
        split_cls.zip_md5 = md5
        return split_cls


class ZippedDataset(IdxDataset):
    default_base_dir = None
    default_zip_filepath = None
    zip_md5 = None

    def __init__(
        self,
        target_dir: Optional[str] = None,
        zip_filepath: Optional[str] = None,
        unzip: bool = True,
        force_unzip: bool = False,
        load: bool = True,
        transpose: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        target_dir : str, default='/tmp/emnist/<split_name>/'
            Directory where all files exist or will be unzipped to (if `unzip` is True).
        zip_filepath : str, default='/tmp/emnist/gzip.zip'
            Filepath to zip file containing all EMNIST split files.
        unzip : bool, default=True
            If True and files don't exist in `target_dir`, unzips all files to `target_dir`.
        force_unzip : bool, default=False
            If True, unzips all files to `target_dir`, even if they exist there.
        load : bool, default=True
            If True, loads data from files in `target_dir`.
        transpose : bool, default=True
            If True, transposes train and test images.
        """

        self.target_dir = (
            os.path.join(self.default_base_dir, type(self).__name__)
            if target_dir is None
            else target_dir
        )

        self.zip_filepath = (
            self.default_zip_filepath if zip_filepath is None else zip_filepath
        )

        self._train_images: Optional[np.ndarray] = None
        self._train_labels: Optional[np.ndarray] = None
        self._test_images: Optional[np.ndarray] = None
        self._test_labels: Optional[np.ndarray] = None

        if unzip or force_unzip:
            self.unzip_files(force_unzip)

        if load:
            self.load(transpose)

    def unzip_files(self, force: bool = False) -> None:
        """
        Unzip files from `zip_filepath` to `target_dir`.

        Parameters
        ----------
        force : bool=False
            If True, unzips all files even if they exist.
        """

        os.makedirs(self.target_dir, exist_ok=True)
        if not check_file_integrity(self.zip_filepath, self.zip_md5):
            raise RuntimeError(
                f"Zip file '{self.zip_filepath}' doesn't exists or its MD5"
                "checksum is not valid. "
                "Use EMNIST(download=True) or emnist.download() to download it"
            )

        for filename, md5 in self.resources.values():
            filepath = os.path.join(self.target_dir, filename)

            if not force and check_file_integrity(filepath, md5):
                continue

            extract_from_zip(self.zip_filepath, filename, self.target_dir)


class NpzDataset(IdxDataset):
    def load(self, transpose=False) -> None:
        """
        Load data from files in `target_dir`.

        Parameters
        ----------
        transpose : bool=False
            If True, transposes train and test images.
        """

        for key, (filename, md5) in self.resources.items():
            filepath = os.path.join(self.target_dir, filename)

            if not check_file_integrity(filepath, md5):
                raise RuntimeError(
                    f"Dataset '{key}' not found in '{filepath}' or MD5 "
                    "checksum is not valid. "
                    "Use download=True or .download() to download it"
                )

            data = np.load(filepath)["arr_0"]
            setattr(self, f"_{key}", data)
